/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>
#include <cstring>
#include <map>
#include <memory>
#include <vector>

#include "common/common.h"
#include "common/context.h"
#include "defs.h"
#include "errors.h"
#include "executor_abstract.h"
#include "parser/ast.h"
#include "record/rm_defs.h"
#include "system/sm_meta.h"

struct group_node {
  std::shared_ptr<RmRecord> tuple;
  std::vector<ColMeta> *group_metas;

  group_node(std::unique_ptr<RmRecord> tuple_, std::vector<ColMeta> *group_metas_)
      : tuple(std::move(tuple_)), group_metas(group_metas_) {}

  bool operator<(const group_node &rhs) const {
    Value left;
    Value right;
    for (auto &col : *group_metas) {
      left.type = col.type;
      left.set_data(tuple->data + col.offset, col.len);

      right.type = col.type;
      right.set_data(rhs.tuple->data + col.offset, col.len);

      int res = left.compare_value(right);

      if (res != 0) {
        return res < 0;
      }
    }

    return false;
  };
};

// 保存每一列的六个个结果。
struct group_result {
  std::vector<Value> mins;
  std::vector<Value> maxs;
  std::vector<Value> sums;
  std::vector<Value> counts;  // 每一列的不含null的count，只不过现在也没有null罢了。
  Value count_star;           // 每一列的含null的count。
  std::vector<Value> values;         // 每一列的原始值。
  const std::vector<ColMeta> *cols;  // 原始元组信息。

  // 进行初始化。
  group_result(RmRecord *tuple, const std::vector<ColMeta> *cols_) : cols(cols_) {
    mins.resize(cols_->size());
    maxs.resize(cols_->size());
    sums.resize(cols_->size());
    counts.resize(cols_->size());
    values.resize(cols_->size());
    count_star.set_int(1);
    for (size_t i = 0; i < cols->size(); ++i) {
      auto &col = cols->at(i);

      mins[i].type = col.type;
      maxs[i].type = col.type;
      sums[i].type = col.type;
      values[i].type = col.type;

      counts[i].set_int(1);

      mins[i].set_data(tuple->data + col.offset, col.len);
      maxs[i].set_data(tuple->data + col.offset, col.len);
      if (col.type != TYPE_STRING) {
        sums[i].set_data(tuple->data + col.offset, col.len);
      }

      values[i].set_data(tuple->data + col.offset, col.len);
      values[i].resize(col.len);
    }
  }

  void update(RmRecord *tuple) {
    Value val;
    count_star.int_val += 1;
    for (size_t i = 0; i < cols->size(); ++i) {
      auto &col = cols->at(i);
      val.type = col.type;
      val.set_data(tuple->data + col.offset, col.len);

      mins[i].min(val);
      maxs[i].max(val);
      if (col.type != TYPE_STRING) {
        sums[i].add(val);
      }

      counts[i].int_val += 1;
    }
  }
};

// 因为having_conds里面可能存在聚合，所以只能在结束统计之后再进行筛选。
// 所以这部分应该可以进行优化，比如：知道没有聚合，就可以提前筛去一些分组。

// 关于这个算子的返回，有一个存在比较大的问题 ：到底该返回什么？
// 在没有order by的情况下，显然，返回select的就行。
// 即使如此，如何指示列的信息呢？ 比如： select a,max(a) from t group by a;
// 这个时候有有两个列都是a，那么在查找的时候怎么抉择？选择那个？
// 必然需要指示聚合函数的信息。
// 即使如此，还可能存在问题： select a,max(a) from t group by a,b order by b;
// 发现，select 和 order by并不重合，但这又是符合结果的。
// 又因为order by在语法树上是该算子的上层，获得不了对应信息。
// 我想到的对策：冗余信息。
// 返回的tuple = select的部分 + group by的部分。
// 然后增加函数：TabCols(),返回查询的表名等信息，方便查找。
// 其实order by可以出现 select a,max(a) from t group by a,b order by a,max(b);
// 但是测试应该不会出现这种情况。
class AggregationAndGroupExecutor : public AbstractExecutor {
 private:
  std::unique_ptr<AbstractExecutor> prev_;
  std::vector<Condition> having_conds_;
  std::vector<ColMeta> group_metas_;  // 需要group的地方的原meta。
  std::vector<ColMeta> tuple_metas_;  // 查询结果的colmeta
  std::map<group_node, group_result> group_map_;
  std::map<group_node, group_result>::iterator tuple_;
  size_t len_;

 public:
  AggregationAndGroupExecutor(std::unique_ptr<AbstractExecutor> prev,
                              const std::vector<TabCol> &group_cols,
                              const std::vector<TabCol> &query_cols,
                              const std::vector<Condition> &having_conds)
      : prev_(std::move(prev)), having_conds_(having_conds) {
    auto &prev_metas = prev_->cols();
    size_t cur_offset = 0;
    for (auto &col : query_cols) {
      if (col.aggregation_type == AGG_COUNT_STAR || col.aggregation_type == AGG_COUNT) {
        tuple_metas_.push_back(ColMeta{col.tab_name, col.col_name, TYPE_INT, 4,
                                       static_cast<int>(cur_offset), false, col.aggregation_type});
        cur_offset += 4;
      } else {
        auto iter =
            std::find_if(prev_metas.begin(), prev_metas.end(), [&col](const ColMeta &col_meta) {
              return col.tab_name == col_meta.tab_name && col.col_name == col_meta.name;
            });

        ColMeta tmp = *iter;
        tmp.offset = cur_offset;
        tmp.aggregation_type = col.aggregation_type;

        tuple_metas_.push_back(tmp);
        cur_offset += iter->len;
      }
    }

    for (auto &col : group_cols) {
      auto iter =
          std::find_if(prev_metas.begin(), prev_metas.end(), [&col](const ColMeta &col_meta) {
            return col.tab_name == col_meta.tab_name && col.col_name == col_meta.name;
          });

      group_metas_.push_back(*iter);

      tuple_metas_.push_back(ColMeta{col.tab_name, col.col_name, iter->type, iter->len,
                                     static_cast<int>(cur_offset), false, col.aggregation_type});
      // .aggregation_type 应该是AGG_NO。

      cur_offset += iter->len;
    }

    len_ = cur_offset;
  }

  void beginTuple() override {
    open_conditions(having_conds_);

    for (prev_->beginTuple(); !prev_->is_end(); prev_->nextTuple()) {
      auto tuple = prev_->Next();
      group_node node(std::move(tuple), &group_metas_);

      auto iter = group_map_.find(node);

      if (iter == group_map_.end()) {
        group_result result(node.tuple.get(), &prev_->cols());
        group_map_.insert(std::make_pair(node, result));
      } else {
        iter->second.update(node.tuple.get());
      }
    }

    // 当且仅当空表，且select全为count。
    // 这个时候才会输出0。否则空表输出空表即可。
    if (group_map_.empty()) {
      int i = static_cast<int>(tuple_metas_.size() - group_metas_.size() - 1);

      for (; i >= 0; --i) {
        auto &col = tuple_metas_[i];
        if (col.aggregation_type != AGG_COUNT && col.aggregation_type != AGG_COUNT_STAR) {
          tuple_ = group_map_.begin();
          return;
        }
      }

      auto tuple = std::make_unique<RmRecord>(prev_->tupleLen());
      group_node node(std::move(tuple), &group_metas_);
      group_result result(node.tuple.get(), &prev_->cols());

      for (auto &count : result.counts) {
        count.int_val = 0;
      }

      result.count_star.int_val = 0;

      group_map_.insert(std::make_pair(node, result));

      tuple_ = group_map_.begin();

      // 这个时候group_map_只有一个值。
      // 不再进行下面的检查，直接置为begin().
      // 这样只会得到一个tuple，
      return;
    }

    for (tuple_ = group_map_.begin(); tuple_ != group_map_.end(); ++tuple_) {
      if (check_having_conditions(tuple_->second)) {
        break;
      }
    }
  }

  void nextTuple() override {
    for (++tuple_; tuple_ != group_map_.end(); ++tuple_) {
      if (check_having_conditions(tuple_->second)) {
        break;
      }
    }
  }

  std::unique_ptr<RmRecord> Next() override {
    auto tuple = std::make_unique<RmRecord>(len_);
    auto &group_result = tuple_->second;
    auto &prev_metas = prev_->cols();

    for (auto &col_meta : tuple_metas_) {
      size_t j =
          std::find_if(prev_metas.begin(), prev_metas.end(),
                       [&col_meta](const ColMeta &col) {
                         return col.tab_name == col_meta.tab_name && col.name == col_meta.name;
                       }) -
          prev_metas.begin();

      switch (col_meta.aggregation_type) {
        case AGG_COUNT: {
          memcpy(tuple->data + col_meta.offset, group_result.counts[j].data(), col_meta.len);
        } break;
        case AGG_COUNT_STAR: {
          memcpy(tuple->data + col_meta.offset, group_result.count_star.data(), col_meta.len);
        } break;
        case AGG_MAX: {
          memcpy(tuple->data + col_meta.offset, group_result.maxs[j].data(), col_meta.len);
        } break;
        case AGG_MIN: {
          memcpy(tuple->data + col_meta.offset, group_result.mins[j].data(), col_meta.len);
        } break;
        case AGG_SUM: {
          memcpy(tuple->data + col_meta.offset, group_result.sums[j].data(), col_meta.len);
        } break;
        case AGG_NO: {
          memcpy(tuple->data + col_meta.offset, group_result.values[j].data(), col_meta.len);
        } break;
      }
    }

    return tuple;
  }

  bool check_aggregation(group_result &result, Condition &cond) {
    auto &prev_metas = prev_->cols();
    size_t i;
    for (i = 0; i < prev_metas.size(); ++i) {
      if (prev_metas[i].name == cond.lhs_col.col_name &&
          prev_metas[i].tab_name == cond.lhs_col.tab_name) {
        break;
      }
    }

    if (i == prev_metas.size() && cond.lhs_col.aggregation_type != AGG_COUNT_STAR) {
      throw InternalError("not found condition.");
    }

    switch (cond.lhs_col.aggregation_type) {
      case AGG_COUNT: {
        auto &val = result.counts[i];
        return val.check_compare(val.compare_value(cond.rhs_val), cond.op);
      } break;
      case AGG_COUNT_STAR: {
        auto &val = result.count_star;
        return val.check_compare(val.compare_value(cond.rhs_val), cond.op);
      } break;
      case AGG_MAX: {
        auto &val = result.maxs[i];
        return val.check_compare(val.compare_value(cond.rhs_val), cond.op);
      }
      case AGG_MIN: {
        auto &val = result.mins[i];
        return val.check_compare(val.compare_value(cond.rhs_val), cond.op);
      }
      case AGG_SUM: {
        auto &val = result.sums[i];
        return val.check_compare(val.compare_value(cond.rhs_val), cond.op);
      }
      case AGG_NO: {
        auto &val = result.values[i];
        return val.check_compare(val.compare_value(cond.rhs_val), cond.op);
      }
    }

    return true;
  }

  bool check_having_conditions(group_result &result) {
    for (auto &cond : having_conds_) {
      if (!check_aggregation(result, cond)) {
        return false;
      }
    }
    return true;
  }

  Rid &rid() override { return _abstract_rid; }
  bool is_end() const override { return tuple_ == group_map_.end(); }
  const std::vector<ColMeta> &cols() const override { return tuple_metas_; }
  size_t tupleLen() const override { return len_; }
};