/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <cstddef>

#include "errors.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "system/sm.h"

class InsertExecutor : public AbstractExecutor {
 private:
  TabMeta tab_;                // 表的元数据
  std::vector<Value> values_;  // 需要插入的数据
  RmFileHandle *fh_;           // 表的数据文件句柄
  std::string tab_name_;       // 表名称
  Rid rid_;  // 插入的位置，由于系统默认插入时不指定位置，因此当前rid_在插入后才赋值
  SmManager *sm_manager_;

 public:
  InsertExecutor(SmManager *sm_manager, const std::string &tab_name, std::vector<Value> values,
                 Context *context) {
    sm_manager_ = sm_manager;
    tab_ = sm_manager_->db_.get_table(tab_name);
    values_ = values;
    tab_name_ = tab_name;
    if (values.size() != tab_.cols.size()) {
      throw InvalidValueCountError();
    }
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    context_ = context;
  };

  std::unique_ptr<RmRecord> Next() override {
    // Make record buffer
    RmRecord rec(fh_->get_file_hdr().record_size);

    // 因为插入暂时只支持一个值还是很好保持原子性的。
    for (size_t i = 0; i < values_.size(); i++) {
      auto &col = tab_.cols[i];
      auto &val = values_[i];

      val.convert_to_type(col.type);
      val.resize(col.len);
      //  val.init_raw(col.len);

      memcpy(rec.data + col.offset, val.data(), col.len);
    }

    // Insert into record file
    rid_ = fh_->insert_record(rec.data, context_);

    // Insert into index
    // 插入时，如果违反唯一性索引，需要抛出异常。
    size_t i;
    for (i = 0; i < tab_.indexes.size(); ++i) {
      auto &index = tab_.indexes[i];
      auto ih =
          sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols))
              .get();
      // char *key = new char[index.col_tot_len]; // 这种也不进行垃圾回收的垃圾写法就别拿出来了。
      RmRecord key(index.col_tot_len);
      index.make_index_tuple(&rec, &key);

      std::vector<Rid> rids;
      ih->get_value(key.data, &rids, context_->txn_);

      // 其中某个索引违反了唯一索引。
      if (!rids.empty()) {
        break;
      }

      ih->insert_entry(key.data, rid_, context_->txn_);
    }

    // 不需要进行回复。
    if (i == tab_.indexes.size()) {
      return nullptr;
    }

    // 进行恢复：之前插入的全部进行删除.
    fh_->delete_record(rid_, context_);  // 表中删除

    // 对索引进行删除恢复。
    for (size_t j = 0; j < i; ++j) {  // 最后一个索引也就是 第i个索引 没有插入，所以不需要进行删除。
      auto &index = tab_.indexes[j];
      auto ih =
          sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols))
              .get();
      // char *key = new char[index.col_tot_len]; // 这种也不进行垃圾回收的垃圾写法就别拿出来了。
      RmRecord key(index.col_tot_len);
      index.make_index_tuple(&rec, &key);

      ih->delete_entry(key.data, context_->txn_);
    }

    throw InternalError("insert_entry has exists");
    return nullptr;
  }
  Rid &rid() override { return rid_; }
  bool is_end() const override { return true; }
};