/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include <cstddef>

#include "defs.h"
#undef NDEBUG

#define private public

#include "record/rm.h"
#include "storage/buffer_pool_manager.h"

#undef private

#include <algorithm>
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <ctime>
#include <iostream>
#include <memory>
#include <random>
#include <set>
#include <string>
#include <thread>  // NOLINT
#include <unordered_map>
#include <vector>

#include "gtest/gtest.h"
#include "index/ix_index_handle.h"
#include "index/ix_manager.h"
#include "replacer/lru_replacer.h"
#include "storage/disk_manager.h"
#include "system/sm_manager.h"

auto disk_manager = std::make_unique<DiskManager>();
auto buffer_pool_manager = std::make_unique<BufferPoolManager>(20, disk_manager.get());
auto rm_manager = std::make_unique<RmManager>(disk_manager.get(), buffer_pool_manager.get());
auto ix_manager = std::make_unique<IxManager>(disk_manager.get(), buffer_pool_manager.get());
auto sm_manager = std::make_unique<SmManager>(disk_manager.get(), buffer_pool_manager.get(),
                                              rm_manager.get(), ix_manager.get());

const std::string TEST_DB_NAME = "B-Plus-Tree-Test-DB";  // 以数据库名作为根目录
class MyTest : public ::testing::Test {
 public:
  // This function is called before every test.
  void SetUp() override {
    if (sm_manager->is_dir(TEST_DB_NAME)) {
      sm_manager->drop_db(TEST_DB_NAME);
    }
    sm_manager->create_db(TEST_DB_NAME);
    sm_manager->open_db(TEST_DB_NAME);
  }

  // This function is called after every test.
  void TearDown() override {
    if (sm_manager->is_dir(TEST_DB_NAME)) {
      sm_manager->drop_db(TEST_DB_NAME);
    }
  };
};

// NOLINTNEXTLINE
TEST_F(MyTest, BPlusTest) {
  std::string table_name = "t";
  std::string col_name = "a";
  std::vector<ColDef> cols_def{ColDef{col_name, TYPE_INT, sizeof(int)}};

  sm_manager->create_table(table_name, cols_def, nullptr);

  std::vector<std::string> index_cols{col_name};
  sm_manager->create_index(table_name, index_cols, nullptr);

  std::string index_name = ix_manager->get_index_name(table_name, index_cols);
  auto index_hdl = sm_manager->ihs_[index_name].get();

  std::random_device seed;

  int n = 1000000;
  int random_cnt = seed() % 10;

  std::cout << "seed : " << random_cnt << std::endl;

  std::vector<int> keys;
  std::vector<Rid> rids;

  for (int i = 1; i <= n; ++i) {
    keys.push_back(i);
  }

  for (int i = 0; i < random_cnt; ++i) {
    std::random_shuffle(keys.begin(), keys.end());
  }
  for (int i : keys) {
    Rid rid{i, i};
    index_hdl->insert_entry((const char *)&i, rid, nullptr);
    // std::cout << "insert : " << i << " successfully" << std::endl;
  }

  std::cout << "insert successfully" << std::endl;

  for (int i = 0; i < random_cnt; ++i) {
    std::random_shuffle(keys.begin(), keys.end());
  }
  for (auto i : keys) {
    // std::cout << "get value : " << i << std::endl;
    index_hdl->get_value((const char *)&i, &rids, nullptr);

    if (rids.empty()) {
      std::cout << "get_value empty : " << i << std::endl;
    }

    if (rids.front().page_no != i) {
      std::cout << "get_value not equal : " << i << std::endl;
    }

    rids.clear();
  }

  std::cout << "query successfully" << std::endl;

  // index_hdl->show_tree();
  for (int i = 0; i < random_cnt; ++i) {
    std::random_shuffle(keys.begin(), keys.end());
  }
  for (auto i : keys) {
    // std::cout << "delete : " << i << std::endl;

    index_hdl->delete_entry((const char *)&i, nullptr);

    index_hdl->get_value((const char *)&i, &rids, nullptr);

    if (!rids.empty()) {
      std::cout << "delete value fails : " << i << std::endl;
    }

    rids.clear();
  }

  std::cout << "delete successfully" << std::endl;
}
